@@ -64,15 +64,17 @@ void read_bookmarks() {
 void write_bookmarks() {
 	struct bookmark *bm;
 	FILE *out;
-	unsigned char *file_name;
+	unsigned char *tmp_file_name, *file_name;
 
-	file_name = stracpy(links_home);
-	if (!file_name) return;
-	add_to_strn(&file_name, "bookmarks");
+	tmp_file_name = stracpy(links_home);
+	if (!tmp_file_name) return;
+	add_to_strn(&tmp_file_name, "bookmarks.tmp");
 	
-	out = fopen(file_name, "w");
-	mem_free(file_name);
-	if (!out) return;
+	out = fopen(tmp_file_name, "w");
+	if (!out) {
+		mem_free(tmp_file_name);
+		return;
+	}
 
 	foreachback(bm, bookmarks) {
 		unsigned char *p = stracpy(bm->title);
@@ -85,7 +87,24 @@ void write_bookmarks() {
 		mem_free(p);
 	}
 
-	fclose(out);
+	if (fclose(out) == EOF) {
+		remove(tmp_file_name);
+		mem_free(tmp_file_name);
+		return;
+	}
+
+	file_name = stracpy(links_home);
+	if (!file_name) {
+		mem_free(tmp_file_name);
+		return;
+	}
+	add_to_strn(&file_name, "bookmarks");
+
+	unlink(file_name);		/* OS/2 needs this */
+	rename(tmp_file_name, file_name);
+	mem_free(tmp_file_name);
+	mem_free(file_name);
+
 }
 
 /* Gets a bookmark by id */
