@@ -4,6 +4,7 @@ void get_system_name()
 {
 	FILE *f;
 	unsigned char *p;
+#ifdef HAVE_POPEN
 	memset(system_name, 0, MAX_STR_LEN);
 	if (!(f = popen("uname -srm", "r"))) goto fail;
 	if (fread(system_name, 1, MAX_STR_LEN - 1, f) <= 0) {
@@ -17,6 +18,7 @@ void get_system_name()
 	}
 	if (system_name[0]) return;
 	fail:
+#endif
 	strcpy(system_name, SYSTEM_NAME);
 }
 
@@ -697,6 +699,11 @@ unsigned char *anonymous_cmd(struct option *o, unsigned char ***argv, int *argc)
 	anonymous = 1;
 	return NULL;
 }
+unsigned char *force_html_cmd(struct option *o, unsigned char ***argv, int *argc)
+{
+	force_html = 1;
+	return NULL;
+}
 unsigned char *dump_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	if (dmp != o->min && dmp) return "Can't use both -dump and -source";
@@ -768,6 +775,9 @@ Options are:\n\
   No local file browsing. No downloads. Executing of viewers\n\
   is allowed, but user can't add or modify entries in\n\
   association table.\n\
+\n\
+ -force-html\n\
+  Treat file as if it had an .html extension.\n\
 \n\
  -dump\n\
   Write a plain-text version of the given HTML document to\n\
@@ -826,6 +836,7 @@ int created_home = 0;
 int no_connect = 0;
 int base_session = 0;
 int dmp = 0;
+int force_html = 0;
 
 int async_lookup = 1;
 int download_utime = 0;
@@ -877,6 +888,7 @@ struct option links_options[] = {
 	1, no_connect_cmd, NULL, NULL, 0, 0, NULL, NULL, "no-connect",
 	1, anonymous_cmd, NULL, NULL, 0, 0, NULL, NULL, "anonymous",
 	1, gen_cmd, num_rd, NULL, 0, MAXINT, &base_session, NULL, "base-session",
+	1, force_html_cmd, NULL, NULL, 0, 0, NULL, NULL, "force-html",
 	1, dump_cmd, NULL, NULL, D_DUMP, 0, NULL, NULL, "dump",
 	1, dump_cmd, NULL, NULL, D_SOURCE, 0, NULL, NULL, "source",
 	1, gen_cmd, num_rd, num_wr, 0, 1, &async_lookup, "async_dns", "async-dns",
