@@ -12,16 +12,17 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 	if (e >= eof || *(e++) != '<') return -1;
 	if (name) *name = e;
 	if (e < eof && *e == '/') e++;
+	if (e >= eof || !isA(*e)) return -1;
 	while (e < eof && isA(*e)) e++;
-	if (e >= eof || (!WHITECHAR(*e) && *e != '>' && *e != '/' && *e != ':')) return -1;
+	if (e >= eof || (!WHITECHAR(*e) && *e != '>' && *e != '<' && *e != '/' && *e != ':')) return -1;
 	if (name && namelen) *namelen = e - *name;
-	while (e < eof && (WHITECHAR(*e) || *e == '/') || *e == ':') e++;
-	if (e >= eof || (!atchr(*e) && *e != '>')) return -1;
+	while (e < eof && (WHITECHAR(*e) || *e == '/' || *e == ':')) e++;
+	if (e >= eof || (!atchr(*e) && *e != '>' && *e != '<')) return -1;
 	if (attr) *attr = e;
 	nextattr:
 	while (e < eof && WHITECHAR(*e)) e++;
-	if (e >= eof || (!atchr(*e) && *e != '>')) return -1;
-	if (*e == '>') goto en;
+	if (e >= eof || (!atchr(*e) && *e != '>' && *e != '<')) return -1;
+	if (*e == '>' || *e == '<') goto en;
 	while (e < eof && atchr(*e)) e++;
 	while (e < eof && WHITECHAR(*e)) e++;
 	if (e >= eof) return -1;
@@ -36,18 +37,18 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 		while (e < eof && *e != uu && *e /*(WHITECHAR(*e) || *e > ' ')*/) e++;
 		if (e >= eof || *e < ' ') return -1;
 		e++;
-		if (e >= eof /*|| (!WHITECHAR(*e) && *e != uu && *e != '>')*/) return -1;
+		if (e >= eof /*|| (!WHITECHAR(*e) && *e != uu && *e != '>' && *e != '<')*/) return -1;
 		if (*e == uu) goto u;
 	} else {
-		while (e < eof && !WHITECHAR(*e) && *e != '>') e++;
+		while (e < eof && !WHITECHAR(*e) && *e != '>' && *e != '<') e++;
 		if (e >= eof) return -1;
 	}
 	while (e < eof && WHITECHAR(*e)) e++;
 	if (e >= eof) return -1;
 	endattr:
-	if (*e != '>') goto nextattr;
+	if (*e != '>' && *e != '<') goto nextattr;
 	en:
-	if (end) *end = e + 1;
+	if (end) *end = e + (*e == '>');
 	return 0;
 }
 
@@ -65,7 +66,7 @@ unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 	int f;
 	aa:
 	while (WHITECHAR(*e)) e++;
-	if (*e == '>') return NULL;
+	if (*e == '>' || *e == '<') return NULL;
 	n = name;
 	while (*n && upcase(*e) == upcase(*n)) e++, n++;
 	f = *n;
@@ -75,7 +76,7 @@ unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 	e++;
 	while (WHITECHAR(*e)) e++;
 	if (!U(*e)) {
-		while (!WHITECHAR(*e) && *e != '>') {
+		while (!WHITECHAR(*e) && *e != '>' && *e != '<') {
 			if (!f) add_chr(a, l, *e);
 			e++;
 		}
@@ -88,8 +89,8 @@ unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 				mem_free(a);
 				return NULL;
 			}
-			if (!f && *e != 10) {
-				if (*e != 9 && *e != 13) add_chr(a, l, *e);
+			if (!f && *e != 13) {
+				if (*e != 9 && *e != 10) add_chr(a, l, *e);
 				else add_chr(a, l, ' ');
 			}
 			e++;
@@ -126,6 +127,7 @@ int has_attr(unsigned char *e, unsigned char *name)
 	return 1;
 }
 
+/*
 unsigned char *get_url_val(unsigned char *e, unsigned char *name)
 {
 	int n = 0;
@@ -140,6 +142,8 @@ unsigned char *get_url_val(unsigned char *e, unsigned char *name)
 	}
 	return pp;
 }
+*/
+#define get_url_val get_attr_val
 
 struct {
 	int n;
@@ -544,12 +548,15 @@ void kill_until(int ls, ...)
 			else if (e->namelen == strlen(s) && !casecmp(e->name, s, strlen(s)))
 				if (!sk) {
 					if (e->dontkill) break;
+					va_end(arg);
 					goto killll;
 				}
 				else if (sk == 1) {
+					va_end(arg);
 					goto killl;
 				} else break;
 		}
+		va_end(arg);
 		if (e->dontkill || (e->namelen == 5 && !casecmp(e->name, "TABLE", 5))) break;
 		if (e->namelen == 2 && upcase(e->name[0]) == 'T' && (upcase(e->name[1]) == 'D' || upcase(e->name[1]) == 'H' || upcase(e->name[1]) == 'R')) break;
 		e = e->next;
@@ -717,7 +724,7 @@ void html_img(unsigned char *a)
 		mem_free(al);
 	}
 	ismap = format.link && has_attr(a, "ismap") && !usemap;
-	if (!(al = get_attr_val(a, "alt")) || !*al) {
+	if ((!(al = get_attr_val(a, "alt")) && !(al = get_attr_val(a, "title"))) || !*al) {
 		if (al) mem_free(al);
 		if (!d_opt->images && !format.link) return;
 		if (usemap) al = stracpy("[USEMAP]");
@@ -865,7 +872,9 @@ void html_hr(unsigned char *a)
 	par_format.rightmargin = margin;
 	if ((i = get_width(a, "width", 1)) == -1) i = par_format.width - 2 * margin - 4;
 	format.attr = AT_GRAPHICS;
+	special_f(ff, SP_NOWRAP, 1);
 	while (i-- > 0) put_chrs(&r, 1, put_chars_f, ff);
+	special_f(ff, SP_NOWRAP, 0);
 	ln_break(2, line_break_f, ff);
 	kill_html_stack_item(&html_top);
 }
@@ -967,13 +976,14 @@ void html_li(unsigned char *a)
 		put_chrs(x, 7, put_chars_f, ff);
 		par_format.leftmargin += 2;
 		par_format.align = AL_LEFT;
+		putsp = -1;
 	} else {
 		char c = 0;
 		char n[32];
 		int t = par_format.flags & P_LISTMASK;
 		int s = get_num(a, "value");
 		if (s != -1) par_format.list_number = s;
-		if ((t != P_roman && t != P_ROMAN && par_format.list_number < 10) || t == P_alpha || t == P_ALPHA) put_chrs(&c, 1, put_chars_f, ff), c = 1;
+		if ((t != P_roman && t != P_ROMAN && par_format.list_number < 10) || t == P_alpha || t == P_ALPHA) put_chrs("&nbsp;", 6, put_chars_f, ff), c = 1;
 		if (t == P_ALPHA || t == P_alpha) {
 			n[0] = par_format.list_number ? (par_format.list_number - 1) % 26 + (t == P_ALPHA ? 'A' : 'a') : 0;
 			n[1] = 0;
@@ -990,6 +1000,7 @@ void html_li(unsigned char *a)
 		par_format.align = AL_LEFT;
 		par_format.list_number = 0;
 		html_top.next->parattr.list_number++;
+		putsp = -1;
 	}
 	line_breax = 2;
 }
@@ -1031,24 +1042,8 @@ void get_html_form(unsigned char *a, struct form *form)
 {
 	unsigned char *al;
 	unsigned char *ch;
-	if ((al = get_url_val(a, "action"))) {
-		char *all = al;
-		while (all[0] == ' ') all++;
-		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
-		form->action = join_urls(format.href_base, all);
-		mem_free(al);
-	} else {
-		if ((ch = strchr(form->action = stracpy(format.href_base), POST_CHAR))) *ch = 0;
-	}
-	if (ch = strchr(form->action, '?')) *ch = 0;
-	if ((al = get_target(a))) {
-		form->target = al;
-	} else {
-		form->target = stracpy(format.target_base);
-	}
 	form->method = FM_GET;
 	if ((al = get_attr_val(a, "method"))) {
-		if (!strcasecmp(al, "get")) form->method = FM_GET;
 		if (!strcasecmp(al, "post")) {
 			char *ax;
 			form->method = FM_POST;
@@ -1060,6 +1055,21 @@ void get_html_form(unsigned char *a, struct form *form)
 		}
 		mem_free(al);
 	}
+	if ((al = get_url_val(a, "action"))) {
+		char *all = al;
+		while (all[0] == ' ') all++;
+		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
+		form->action = join_urls(format.href_base, all);
+		mem_free(al);
+	} else {
+		if ((ch = strchr(form->action = stracpy(format.href_base), POST_CHAR))) *ch = 0;
+		if (form->method == FM_GET && (ch = strchr(form->action, '?'))) *ch = 0;
+	}
+	if ((al = get_target(a))) {
+		form->target = al;
+	} else {
+		form->target = stracpy(format.target_base);
+	}
 	form->num = a - startf;
 }
 
@@ -1069,6 +1079,8 @@ void find_form_for_input(unsigned char *i)
 	int namelen;
 	if (form.action) mem_free(form.action);
 	if (form.target) mem_free(form.target);
+	memset(&form, 0, sizeof(form));
+	if (!special_f(ff, SP_USED, NULL)) return;
 	if (last_input_tag && i <= last_input_tag && i > last_form_tag) {
 		get_html_form(last_form_attr, &form);
 		return;
@@ -1196,6 +1208,8 @@ void html_input(unsigned char *a)
 	if (fc->type != FC_FILE) fc->default_value = get_attr_val(a, "value");
 	if (fc->type == FC_CHECKBOX && !fc->default_value) fc->default_value = stracpy("on");
 	if ((fc->size = get_num(a, "size")) == -1) fc->size = HTML_DEFAULT_INPUT_SIZE;
+	fc->size++;
+	if (fc->size > d_opt->xw) fc->size = d_opt->xw;
 	if ((fc->maxlength = get_num(a, "maxlength")) == -1) fc->maxlength = MAXINT;
 	if (fc->type == FC_CHECKBOX || fc->type == FC_RADIO) fc->default_state = has_attr(a, "checked");
 	fc->ro = has_attr(a, "disabled") ? 2 : has_attr(a, "readonly") ? 1 : 0;
@@ -1276,10 +1290,10 @@ void html_option(unsigned char *a)
 			internal("parse element failed");
 			goto x;
 		}
-		r:
+		rrrr:
 		while (p < eoff && WHITECHAR(*p)) p++;
 		while (p < eoff && !WHITECHAR(*p) && *p != '<') {
-			p:
+			pppp:
 			add_chr_to_str(&val, &l, *p), p++;
 		}
 		r = p;
@@ -1287,15 +1301,15 @@ void html_option(unsigned char *a)
 		if (r >= eoff) goto x;
 		if (r - 2 <= eoff && (r[1] == '!' || r[1] == '?')) {
 			p = skip_comment(r, eoff);
-			goto r;
+			goto rrrr;
 		}
-		if (parse_element(r, eoff, &name, &namelen, NULL, &p)) goto p;
+		if (parse_element(r, eoff, &name, &namelen, NULL, &p)) goto pppp;
 		if (!((namelen == 6 && !casecmp(name, "OPTION", 6)) ||
 		    (namelen == 7 && !casecmp(name, "/OPTION", 7)) ||
 		    (namelen == 6 && !casecmp(name, "SELECT", 6)) ||
 		    (namelen == 7 && !casecmp(name, "/SELECT", 7)) ||
 		    (namelen == 8 && !casecmp(name, "OPTGROUP", 8)) ||
-		    (namelen == 9 && !casecmp(name, "/OPTGROUP", 9)))) goto r;
+		    (namelen == 9 && !casecmp(name, "/OPTGROUP", 9)))) goto rrrr;
 	}
 	x:
 	fc->form_num = last_form_tag - startf;
@@ -1575,6 +1589,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	fc->name = get_attr_val(attr, "name");
 	fc->type = FC_SELECT;
 	fc->default_state = preselect < 0 ? 0 : preselect;
+	fc->default_value = order ? stracpy(val[fc->default_state]) : stracpy("");
 	fc->ro = has_attr(attr, "disabled") ? 2 : has_attr(attr, "readonly") ? 1 : 0;
 	fc->nvalues = order;
 	fc->values = val;
@@ -1633,8 +1648,11 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	for (p = fc->default_value; p && p[0]; p++) if (p[0] == '\r')
 		if (p[1] == '\n' || (p > fc->default_value && p[-1] == '\n')) memcpy(p, p + 1, strlen(p)), p--;
 		else p[0] = '\n';
-	if ((cols = get_num(attr, "cols")) <= 0) cols = 20;
+	if ((cols = get_num(attr, "cols")) <= 0) cols = HTML_DEFAULT_INPUT_SIZE;
+	cols++;
 	if ((rows = get_num(attr, "rows")) <= 0) rows = 1;
+	if (cols > d_opt->xw) cols = d_opt->xw;
+	if (rows > d_opt->yw) rows = d_opt->yw;
 	fc->cols = cols;
 	fc->rows = rows;
 	fc->wrap = has_attr(attr, "wrap");
@@ -1919,6 +1937,7 @@ struct element_info elements[] = {
 	{"BLOCKQUOTE",	html_blockquote,2, 0},
 	{"ADDRESS",	html_address,	2, 0},
 	{"PRE",		html_pre,	2, 0},
+	{"LISTING",	html_pre,	2, 0},
 
 	{"UL",		html_ul,	1, 0},
 	{"DIR",		html_ul,	1, 0},
@@ -2369,15 +2388,16 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 		clr_spaces(*title);
 		goto ps;
 	}
-	if (namelen != 4 && !casecmp(name, "META", 4)) goto se;
+	if (namelen != 4 || casecmp(name, "META", 4)) goto se;
+	if ((he = get_attr_val(attr, "charset"))) {
+		add_to_str(head, hdl, "Charset: ");
+		add_to_str(head, hdl, he);
+		mem_free(he);
+	}
 	if (!(he = get_attr_val(attr, "http-equiv"))) goto se;
 	c = get_attr_val(attr, "content");
 	add_to_str(head, hdl, he);
 	if (c) add_to_str(head, hdl, ": "), add_to_str(head, hdl, c), mem_free(c);
-	if (!strcasecmp(he, "content-type")) {
-		c = get_attr_val(attr, "charset");
-		if (c) add_to_str(head, hdl, ";charset="), add_to_str(head, hdl, c), mem_free(c);
-	}
 	mem_free(he);
 	add_to_str(head, hdl, "\r\n");
 	goto se;
