@@ -129,6 +129,8 @@ static inline int fg_color(int fg, int bg)
 
 #define ALIGN(x) (((x)+0x7f)&~0x7f)
 
+int nowrap = 0;
+
 static inline int xpand_lines(struct part *p, int y)
 {
 	/*if (y >= p->y) p->y = y + 1;*/
@@ -500,6 +502,7 @@ void put_chars(struct part *p, unsigned char *c, int l)
 	end_format_change:
 	if (p->cx == par_format.leftmargin && *c == ' ' && par_format.align != AL_NO) c++, l--;
 	if (p->y < p->cy + 1) p->y = p->cy + 1;
+	if (nowrap && p->cx + l > rm(par_format)) return;
 	set_hline(p, p->cx, p->cy, l, c, (((fg&0x08)<<3)|(bg<<3)|(fg&0x07))<<8, 1);
 	p->cx += l;
 	nobreak = 0;
@@ -654,6 +657,12 @@ void html_form_control(struct part *p, struct form_control *fc)
 			mem_free(fc->default_value);
 			fc->default_value = dv;
 		}
+		/*
+		for (i = 0; i < fc->nvalues; i++) if ((dv = convert_string(convert_table, fc->values[i], strlen(fc->values[i])))) {
+			mem_free(fc->values[i]);
+			fc->values[i] = dv;
+		}
+		*/
 	}
 	add_to_list(p->data->forms, fc);
 }
@@ -706,25 +715,36 @@ void *html_special(struct part *p, int c, ...)
 	switch (c) {
 		case SP_TAG:
 			t = va_arg(l, unsigned char *);
+			va_end(l);
 			html_tag(p->data, t, X(p->cx), Y(p->cy));
 			break;
 		case SP_CONTROL:
 			fc = va_arg(l, struct form_control *);
+			va_end(l);
 			html_form_control(p, fc);
 			break;
 		case SP_TABLE:
+			va_end(l);
 			return convert_table;
 		case SP_USED:
+			va_end(l);
 			return (void *)!!p->data;
 		case SP_FRAMESET:
 			fsp = va_arg(l, struct frameset_param *);
+			va_end(l);
 			return create_frameset(p->data, fsp);
 		case SP_FRAME:
 			fp = va_arg(l, struct frame_param *);
+			va_end(l);
 			create_frame(fp);
 			break;
+		case SP_NOWRAP:
+			nowrap = va_arg(l, int);
+			va_end(l);
+			break;
 		default:
 			internal("html_special: unknown code %d", c);
+			va_end(l);
 	}
 	return NULL;
 }
@@ -737,6 +757,26 @@ void do_format(char *start, char *end, struct part *part, unsigned char *head)
 
 int margin;
 
+struct table_cache_entry {
+	struct table_cache_entry *next;
+	struct table_cache_entry *prev;
+	unsigned char *start;
+	unsigned char *end;
+	int align;
+	int m;
+	int width;
+	int xs;
+	int link_num;
+	struct part p;
+};
+
+struct list_head table_cache = { &table_cache, &table_cache };
+
+void free_table_cache()
+{
+	free_list(table_cache);
+}
+
 struct part *format_html_part(unsigned char *start, unsigned char *end, int align, int m, int width, struct f_data *data, int xs, int ys, unsigned char *head, int link_num)
 {
 	struct part *p;
@@ -747,6 +787,15 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	int lm = margin;
 	int ef = empty_format;
 	struct form_control *fc;
+	struct table_cache_entry *tce;
+	if (!data) foreach(tce, table_cache) {
+		if (tce->start == start && tce->end == end && tce->align == align && tce->m == m && tce->width == width && tce->xs == xs && tce->link_num == link_num) {
+			if ((p = mem_alloc(sizeof(struct part)))) {
+				memcpy(p, &tce->p, sizeof(struct part));
+				return p;
+			}
+		}
+	}
 	if (ys < 0) {
 		internal("format_html_part: ys == %d", ys);
 		return NULL;
@@ -796,6 +845,7 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	p->cx = -1;
 	p->cy = 0;
 	do_format(start, end, p, head);
+	if (p->xmax < p->x) p->xmax = p->x;
 	nobreak = 0;
 	line_breax = 1;
 	if (last_link) mem_free(last_link);
@@ -825,6 +875,17 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	empty_format = ef;
 	last_link = last_image = last_target = NULL;
 	last_form = NULL;
+	if (table_level > 1 && !data && ((tce = mem_alloc(sizeof(struct table_cache_entry))))) {
+		tce->start = start;
+		tce->end = end;
+		tce->align = align;
+		tce->m = m;
+		tce->width = width;
+		tce->xs = xs;
+		tce->link_num = link_num;
+		memcpy(&tce->p, p, sizeof(struct part));
+		add_to_list(table_cache, tce);
+	}
 	return p;
 }
 
