@@ -1,3 +1,7 @@
+#ifndef __EXTENSIONS__
+#define __EXTENSIONS__
+#endif
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -51,6 +55,20 @@
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+#ifdef HAVE_SYS_CYGWIN_H
+#include <sys/cygwin.h>
+#endif
+#ifdef HAVE_IO_H
+#include <io.h>
+#endif
+/*
+#ifdef HAVE_WAIT_H
+#include <wait.h>
+#endif
+*/
 
 #ifdef HAVE_NETINET_IN_SYSTM_H
 #include <netinet/in_systm.h>
@@ -103,6 +121,10 @@ static inline int getpgrp()
 #endif
 */
 
+#define option option_dirty_workaround_for_name_clash_with_include_on_cygwin
+#define table table_dirty_workaround_for_name_clash_with_libraries_on_macos
+#define scroll scroll_dirty_workaround_for_name_clash_with_libraries_on_macos
+
 /* error.c */
 
 void do_not_optimize_here(void *p);
@@ -364,14 +386,14 @@ static inline long strtolx(unsigned char *c, unsigned char **end)
 	if (!*end) return l;
 	if (upcase(**end) == 'K') {
 		(*end)++;
-		if (l < LONG_MIN / 1024) return LONG_MIN;
-		if (l > LONG_MAX / 1024) return LONG_MAX;
+		if (l < -MAXINT / 1024) return -MAXINT;
+		if (l > MAXINT / 1024) return MAXINT;
 		return l * 1024;
 	}
 	if (upcase(**end) == 'M') {
 		(*end)++;
-		if (l < LONG_MIN / (1024 * 1024)) return LONG_MIN;
-		if (l > LONG_MAX / (1024 * 1024)) return LONG_MAX;
+		if (l < -MAXINT / (1024 * 1024)) return -MAXINT;
+		if (l > MAXINT / (1024 * 1024)) return MAXINT;
 		return l * (1024 * 1024);
 	}
 	return l;
@@ -522,6 +544,7 @@ int resize_window(int, int);
 int can_resize_window(int);
 int can_open_os_shell(int);
 struct open_in_new *get_open_in_new(int);
+void set_highpri();
 
 /* select.c */
 
@@ -678,10 +701,11 @@ static inline int getpri(struct connection *c)
 #define S_CONN		2
 #define S_SSL_NEG	3
 #define S_SENT		4
-#define S_GETH		5
-#define S_PROC		6
-#define S_TRANS		7
-#define S_QUESTIONS	8
+#define S_LOGIN		5
+#define S_GETH		6
+#define S_PROC		7
+#define S_TRANS		8
+#define S_QUESTIONS	9
 
 #define S_WAIT_REDIR		-999
 #define S_OK			-1000
@@ -950,6 +974,7 @@ struct window {
 #define ENV_OS2VIO	4
 #define ENV_BE		8
 #define ENV_TWIN	16
+#define ENV_WIN32	32
 
 struct terminal {
 	struct terminal *next;
@@ -998,6 +1023,7 @@ struct term_spec {
 #define ATTR_FRAME	0x8000
 
 extern struct list_head term_specs;
+extern struct list_head terminals;
 
 int hard_write(int, unsigned char *, int);
 int hard_read(int, unsigned char *, int);
@@ -1474,6 +1500,7 @@ struct session {
 
 extern struct list_head sessions;
 
+time_t parse_http_date(const char *);
 unsigned char *encode_url(unsigned char *);
 unsigned char *decode_url(unsigned char *);
 unsigned char *subst_file(unsigned char *, unsigned char *);
@@ -1866,6 +1893,7 @@ void scan_http_equiv(unsigned char *, unsigned char *, unsigned char **, int *,
 #define SP_USED		3
 #define SP_FRAMESET	4
 #define SP_FRAME	5
+#define SP_NOWRAP	6
 
 struct frameset_param {
 	struct frameset_desc *parent;
@@ -1912,6 +1940,8 @@ void xset_hchar(struct part *, int, int, unsigned);
 void xset_hchars(struct part *, int, int, int, unsigned);
 void align_line(struct part *, int);
 
+void free_table_cache();
+
 struct conv_table *get_convert_table(unsigned char *, int, int, int *, int *, int);
 extern int format_cache_entries;
 long formatted_info(int);
@@ -1965,6 +1995,7 @@ extern int created_home;
 
 extern int no_connect;
 extern int base_session;
+extern int force_html;
 
 #define D_DUMP		1
 #define D_SOURCE	2
