@@ -128,40 +128,38 @@ void unhandle_basic_signals(struct terminal *term)
 #endif
 }
 
+int terminal_pipe[2];
+
 int attach_terminal(int in, int out, int ctl, void *info, int len)
 {
 	struct terminal *term;
-	int fd[2];
-	if (c_pipe(fd)) {
-		mem_free(info);
-		error("ERROR: can't create pipe for internal communication");
-		return -1;
-	}
-	fcntl(fd[0], F_SETFL, O_NONBLOCK);
-	fcntl(fd[1], F_SETFL, O_NONBLOCK);
-	handle_trm(in, out, out, fd[1], ctl, info, len);
+	fcntl(terminal_pipe[0], F_SETFL, O_NONBLOCK);
+	fcntl(terminal_pipe[1], F_SETFL, O_NONBLOCK);
+	handle_trm(in, out, out, terminal_pipe[1], ctl, info, len);
 	mem_free(info);
-	if ((term = init_term(fd[0], out, win_func))) {
+	if ((term = init_term(terminal_pipe[0], out, win_func))) {
 		handle_basic_signals(term);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
-		return fd[1];
+		return terminal_pipe[1];
 	}
-	close(fd[0]);
-	close(fd[1]);
+	close(terminal_pipe[0]);
+	close(terminal_pipe[1]);
 	return -1;
 }
 
 struct status dump_stat;
 int dump_pos;
 
+int dump_red_count = 0;
+
 void end_dump(struct status *stat, void *p)
 {
 	struct cache_entry *ce = stat->ce;
 	int oh = get_output_handle();
 	if (oh == -1) return;
-	if (ce && ce->redirect) {
+	if (ce && ce->redirect && dump_red_count++ < MAX_REDIRECTS) {
 		unsigned char *u, *p;
 		if (stat->state >= 0) change_connection(stat, NULL, PRI_CANCEL);
-		u = stracpy(ce->redirect);
+		u = join_urls(ce->url, ce->redirect);
 		if (!http_bugs.bug_302_redirect) if (!ce->redirect_get && (p = strchr(ce->url, POST_CHAR))) add_to_strn(&u, p);
 		load_url(u, stat, PRI_MAIN, 0);
 		mem_free(u);
@@ -245,8 +243,16 @@ void init()
 	set_sigcld();
 	init_home();
 	init_keymaps();
+
+/* OS/2 has some stupid bug and the pipe must be created before socket :-/ */
+	if (c_pipe(terminal_pipe)) {
+		error("ERROR: can't create pipe for internal communication");
+		goto ttt;
+	}
 	if (!(u = parse_options(ac - 1, av + 1))) goto ttt;
 	if (!no_connect && (uh = bind_to_af_unix()) != -1) {
+		close(terminal_pipe[0]);
+		close(terminal_pipe[1]);
 		if (!(info = create_session_info(base_session, u, &len))) goto ttt;
 		handle_trm(get_input_handle(), get_output_handle(), uh, uh, get_ctl_handle(), info, len);
 		handle_basic_signals(NULL);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
@@ -273,6 +279,8 @@ void init()
 		}
 	} else {
 		unsigned char *uu, *wd;
+		close(terminal_pipe[0]);
+		close(terminal_pipe[1]);
 		if (!*u) {
 			fprintf(stderr, "URL expected after %s\n.", dmp == D_DUMP ? "-dump" : "-source");
 			goto ttt;
