@@ -4,7 +4,11 @@
 #include <sys/ioctl.h>
 #endif
 
-#ifdef HAVE_LIBGPM
+#if defined(HAVE_LIBGPM) && defined(HAVE_GPM_H)
+#define USE_GPM
+#endif
+
+#ifdef USE_GPM
 #include <gpm.h>
 #endif
 
@@ -55,7 +59,7 @@ int get_e(char *env)
 
 /* Terminal size */
 
-#if defined(UNIX) || defined(BEOS)
+#if defined(UNIX) || defined(BEOS) || defined(RISCOS) || defined(WIN32)
 
 void sigwinch(void *s)
 {
@@ -80,8 +84,11 @@ int get_terminal_size(int fd, int *x, int *y)
 		if (!(*x = ws.ws_col) && !(*x = get_e("COLUMNS"))) *x = 80;
 		if (!(*y = ws.ws_row) && !(*y = get_e("LINES"))) *y = 24;
 		return 0;
+	} else {
+		!(*x = get_e("COLUMNS")) && (*x = 80);
+		!(*y = get_e("LINES")) && (*y = 24);
 	}
-	return -1;
+	return 0;
 }
 
 #elif defined(OS2)
@@ -108,6 +115,7 @@ void winch_thread(void)
 	static int old_xsize, old_ysize;
 	static int cur_xsize, cur_ysize;
 
+	signal(SIGPIPE, SIG_IGN);
 	if (get_terminal_size(0, &old_xsize, &old_ysize)) return;
 	while (1) {
 		if (get_terminal_size(0, &cur_xsize, &cur_ysize)) return;
@@ -200,7 +208,7 @@ int get_terminal_size(int fd, int *x, int *y)
 
 /* Pipe */
 
-#if defined(UNIX) || defined(BEOS)
+#if defined(UNIX) || defined(BEOS) || defined(RISCOS)
 
 void set_bin(int fd)
 {
@@ -236,12 +244,20 @@ int check_file_name(unsigned char *file)
 
 /* Exec */
 
+int can_twterm() /* Check if it make sense to call a twterm. */
+{
+	static int xt = -1;
+	if (xt == -1) xt = !!getenv("TWDISPLAY");
+	return xt;
+}
+
+
 #if defined(UNIX) || defined(WIN32)
 
 int is_xterm()
 {
 	static int xt = -1;
-	if (xt == -1) xt = !!getenv("DISPLAY");
+	if (xt == -1) xt = getenv("DISPLAY") && *getenv("DISPLAY");
 	return xt;
 }
 
@@ -252,11 +268,18 @@ int is_xterm()
 	return 0;
 }
 
+#elif defined(RISCOS)
+
+int is_xterm()
+{
+       return 1;
+}
+
 #endif
 
 tcount resize_count = 0;
 
-#if defined(UNIX) || defined(WIN32) || defined(BEOS)
+#if defined(UNIX) || defined(WIN32) || defined(BEOS) || defined(RISCOS)
 
 #if defined(BEOS) && defined(HAVE_SETPGID)
 
@@ -274,6 +297,28 @@ int exe(char *path)
 	return 0;
 }
 
+#elif defined(WIN32)
+
+int exe(char *path)
+{
+	int r;
+	unsigned char *x1 = !GETSHELL ? DEFAULT_SHELL : GETSHELL;
+	unsigned char *x = *path != '"' ? " /c start /wait " : " /c start /wait \"\" ";
+	unsigned char *p = malloc((strlen(x1) + strlen(x) + strlen(path)) * 2 + 1);
+	if (!p) return -1;
+	strcpy(p, x1);
+	strcat(p, x);
+	strcat(p, path);
+	x = p;
+	while (*x) {
+		if (*x == '\\') memmove(x + 1, x, strlen(x) + 1), x++;
+		x++;
+	}
+	r = system(p);
+	free(p);
+	return r;
+}
+
 #else
 
 int exe(char *path)
@@ -621,7 +666,7 @@ void *bgpt(struct tdata *t)
 
 #endif
 
-#if defined(UNIX) || defined(OS2)
+#if defined(UNIX) || defined(OS2) || defined(RISCOS)
 
 void terminate_osdep() {}
 
@@ -744,10 +789,11 @@ int start_thread(void (*fn)(void *, int), void *ptr, int l)
 int tp = -1;
 int ti = -1;
 
-void input_thread(void *p)	/* !!! FIXME: race-condition with mouse */
+void input_thread(void *p)
 {
 	char c[2];
 	int h = (int)p;
+	signal(SIGPIPE, SIG_IGN);
 	while (1) {
 		/*c[0] = _read_kbd(0, 1, 1);
 		if (c[0]) if (write(h, c, 1) <= 0) break;
@@ -775,7 +821,7 @@ void input_thread(void *p)	/* !!! FIXME: race-condition with mouse */
 }
 #endif /* #ifdef HAVE_READ_KBD */
 
-#if defined(HAVE_MOUOPEN) && !defined(HAVE_LIBGPM)
+#if defined(HAVE_MOUOPEN) && !defined(USE_GPM)
 
 #define USING_OS2_MOUSE
 
@@ -791,6 +837,7 @@ struct os2_mouse_spec {
 	void *data;
 	unsigned char buffer[sizeof(struct event)];
 	int bufptr;
+	int terminate;
 };
 
 void mouse_thread(void *p)
@@ -802,6 +849,7 @@ void mouse_thread(void *p)
 	A_DECL(USHORT, rd);
 	A_DECL(USHORT, mask);
 	struct event ev;
+	signal(SIGPIPE, SIG_IGN);
 	ev.ev = EV_MOUSE;
 	if (MouOpen(NULL, mh)) goto ret;
 	mouse_h = *mh;
@@ -818,7 +866,7 @@ void mouse_thread(void *p)
 #ifdef HAVE_SYS_FMUTEX_H
 		_fmutex_request(&mouse_mutex, _FMR_IGNINT);
 #endif
-		MouDrawPtr(*mh);
+		if (!oms->terminate) MouDrawPtr(*mh);
 #ifdef HAVE_SYS_FMUTEX_H
 		_fmutex_release(&mouse_mutex);
 #endif
@@ -880,6 +928,7 @@ void *handle_mouse(int cons, void (*fn)(void *, unsigned char *, int), void *dat
 	oms->fn = fn;
 	oms->data = data;
 	oms->bufptr = 0;
+	oms->terminate = 0;
 	if (c_pipe(oms->p)) {
 		free(oms);
 		return NULL;
@@ -892,8 +941,11 @@ void *handle_mouse(int cons, void (*fn)(void *, unsigned char *, int), void *dat
 void unhandle_mouse(void *om)
 {
 	struct os2_mouse_spec *oms = om;
+	want_draw();
+	oms->terminate = 1;
 	set_handlers(oms->p[0], NULL, NULL, NULL, NULL);
 	close(oms->p[0]);
+	done_draw();
 }
 
 void want_draw()
@@ -1074,7 +1126,7 @@ int get_input_handle()
 	tp = fd[1];
 	_beginthread(input_thread, NULL, 0x10000, (void *)tp);
 /*
-#if defined(HAVE_MOUOPEN) && !defined(HAVE_LIBGPM)
+#if defined(HAVE_MOUOPEN) && !defined(USE_GPM)
 	_beginthread(mouse_thread, NULL, 0x10000, (void *)tp);
 #endif
 */
@@ -1083,6 +1135,9 @@ int get_input_handle()
 
 #elif defined(WIN32)
 
+void input_function (int fd);
+void set_proc_id (int id);
+
 int get_input_handle()
 {
 	int	fd[2] ;
@@ -1125,7 +1180,7 @@ void cfmakeraw(struct termios *t)
 }
 #endif
 
-#ifdef HAVE_LIBGPM
+#ifdef USE_GPM
 
 struct gpm_mouse_spec {
 	int h;
@@ -1186,7 +1241,7 @@ void unhandle_mouse(void *h)
 void *handle_mouse(int cons, void (*fn)(void *, unsigned char *, int), void *data) { return NULL; }
 void unhandle_mouse(void *data) { }
 
-#endif /* #ifdef HAVE_LIBGPM */
+#endif /* #ifdef USE_GPM */
 
 #if defined(OS2)
 
@@ -1205,6 +1260,13 @@ int get_system_env()
 	return 0;
 }
 
+#elif defined(WIN32)
+
+int get_system_env()
+{
+	return ENV_WIN32;
+}
+
 #else
 
 int get_system_env()
@@ -1223,6 +1285,13 @@ void exec_new_links(struct terminal *term, unsigned char *xterm, unsigned char *
 	mem_free(str);
 }
 
+void open_in_new_twterm(struct terminal *term, unsigned char *exe, unsigned char *param)
+{
+	unsigned char *twterm;
+	if (!(twterm = getenv("LINKS_TWTERM"))) twterm = "twterm -e";
+	exec_new_links(term, twterm, exe, param);
+}
+
 void open_in_new_xterm(struct terminal *term, unsigned char *exe, unsigned char *param)
 {
 	unsigned char *xterm;
@@ -1247,6 +1316,13 @@ void open_in_new_fullscreen(struct terminal *term, unsigned char *exe, unsigned
 }
 #endif
 
+#ifdef WIN32
+void open_in_new_win32(struct terminal *term, unsigned char *exe, unsigned char *param)
+{
+	exec_new_links(term, "", exe, param);
+}
+#endif
+
 #ifdef BEOS
 void open_in_new_be(struct terminal *term, unsigned char *exe, unsigned char *param)
 {
@@ -1261,11 +1337,15 @@ struct {
 	unsigned char *hk;
 } oinw[] = {
 	ENV_XWIN, open_in_new_xterm, TEXT(T_XTERM), TEXT(T_HK_XTERM),
+	ENV_TWIN, open_in_new_twterm, TEXT(T_TWTERM), TEXT(T_HK_TWTERM),
 	ENV_SCREEN, open_in_new_screen, TEXT(T_SCREEN), TEXT(T_HK_SCREEN),
 #ifdef OS2
 	ENV_OS2VIO, open_in_new_vio, TEXT(T_WINDOW), TEXT(T_HK_WINDOW),
 	ENV_OS2VIO, open_in_new_fullscreen, TEXT(T_FULL_SCREEN), TEXT(T_HK_FULL_SCREEN),
 #endif
+#ifdef WIN32
+	ENV_WIN32, open_in_new_win32, TEXT(T_WINDOW), TEXT(T_HK_WINDOW),
+#endif
 #ifdef BEOS
 	ENV_BE, open_in_new_be, TEXT(T_BEOS_TERMINAL), TEXT(T_HK_BEOS_TERMINAL),
 #endif
@@ -1306,3 +1386,14 @@ int can_open_os_shell(int environment)
 #endif
 	return 1;
 }
+
+#ifndef OS2
+void set_highpri()
+{
+}
+#else
+void set_highpri()
+{
+	DosSetPriority(PRTYS_PROCESS, PRTYC_FOREGROUNDSERVER, 0, 0);
+}
+#endif
