@@ -257,6 +257,7 @@ void ses_imgmap(struct session *ses)
 		internal("can't find cache entry");
 		return;
 	}
+	defrag_entry(ce);
 	fr = ce->frag.next;
 	if ((void *)fr == &ce->frag) return;
 	if (get_image_map(ce->head, fr->data, fr->data + fr->length, ses->goto_position, &menu, &ml, ses->imgmap_href_base, ses->imgmap_target_base, ses->term->spec->charset, ses->ds.assume_cp, ses->ds.hard_assume))
@@ -420,7 +421,7 @@ void download_window_function(struct dialog_data *dlg)
 	x = dlg->x + DIALOG_LB;
 	dlg_format_text(term, term, u, x, &y, w, NULL, COLOR_DIALOG_TEXT, AL_LEFT);
 	if (t && stat->prg->size >= 0) {
-		unsigned char q[5];
+		unsigned char q[64];
 		int p = w - 6;
 		y++;
 		set_only_char(term, x, y, '[');
@@ -650,7 +651,17 @@ unsigned char *subst_file(unsigned char *prog, unsigned char *file)
 		add_bytes_to_str(&n, &l, prog, p);
 		prog += p;
 		if (*prog == '%') {
+#if defined(HAVE_CYGWIN_CONV_TO_FULL_WIN32_PATH)
+#ifdef MAX_PATH
+			unsigned char new_path[MAX_PATH];
+#else
+			unsigned char new_path[1024];
+#endif
+			cygwin_conv_to_full_win32_path(file, new_path);
+			add_to_str(&n, &l, new_path);
+#else
 			add_to_str(&n, &l, file);
+#endif
 			prog++;
 		}
 	}
@@ -976,12 +987,15 @@ void request_frame(struct session *ses, unsigned char *name, unsigned char *uurl
 	url = stracpy(uurl);
 	pos = extract_position(url);
 	if (!(frm = mem_alloc(sizeof(struct frame) + strlen(url) + 1))) {
-		mem_free(url); if (pos) mem_free(pos);
+		mem_free(url);
+		if (pos) mem_free(pos);
 		return;
 	}
 	memset(frm, 0, sizeof(struct frame));
 	if (!(frm->name = stracpy(name))) {
 		mem_free(frm);
+		mem_free(url);
+		if (pos) mem_free(pos);
 		return;
 	}
 	init_vs(&frm->vs, url);
