@@ -10,6 +10,7 @@ struct {
 	int need_slash_after_host;
 } protocols[]= {
 		{"file", 0, file_func, NULL, 1, 1, 0},
+		{"https", 443, https_func, NULL, 0, 1, 1},
 		{"http", 80, http_func, NULL, 0, 1, 1},
 		{"proxy", 3128, proxy_func, NULL, 0, 1, 1},
 		{"ftp", 21, ftp_func, NULL, 0, 1, 1},
@@ -152,6 +153,14 @@ unsigned char *get_pass(unsigned char *url)
 	return memacpy(h, hl);
 }
 
+unsigned char *get_port_str(unsigned char *url)
+{
+	unsigned char *h;
+	int hl;
+	if (parse_url(url, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &h, &hl, NULL, NULL, NULL)) return NULL;
+	return hl ? memacpy(h, hl) : NULL;
+}
+
 int get_port(unsigned char *url)
 {
 	unsigned char *h;
